//////////////
// Topology //
//////////////
//
//            startX         startX+roadLength
//     startY ---------------
//               lane 0
//     startY ---------------
// +laneWidth    lane 1
//            ---------------
//               ...
//     startY ---------------
//    +(N-1)*    lane N
//  laneWidth ---------------

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "random.c"
#include <random>
#include <cmath>
#include <set>
#define OUT_X 1000000.0
#define OUT_Y 1000000.0
#define MAX_SPEED 300000000.0
#define MAX_DOUBLE 1e20
#define FILENAME 40
#define MAXDECEL 8
#define V1 -1
#define V2 -1
typedef enum {FALSE, TRUE} boolean;

#define DEBUG 0


double stmu_lane[] = {2.5,2.2,2,1.75};
double stsigma_lane[] = {0.75,0.62,0.55,0.45};

class Lane;

std::random_device rd;
std::mt19937 gen(rd());
std::default_random_engine generator;
std::set<int> maxDecelVSet;
////////////////////
// Node (vehicle) //
////////////////////
class Node {
	public:
		Node(int id, double length, Lane *lane, double x, double y) {
			id_ = id;
			length_ = length;
			lane_ = lane;
			x_ = x;
			y_ = y;
			alreadyTraversed_=0;
			speed_ = 0.0;
			targetSpeed_ = 0.0;
			targetSpeedChangeCount_ = 0;
			dv = 0.0;
			next_ = NULL;
			prev_ = NULL;
			safeTime_=0;

		}

		int& id() { return(id_); }
		double& length() { return(length_); }
		Lane*& lane() { return(lane_); }
		double& x() { return(x_); }
		double& y() { return(y_); }
		double& get_dv() {return(dv);}
		int& alreadyTraversed() {return(alreadyTraversed_);}
		double& safeTime() {return(safeTime_);}
		// IDM methods
		double& speed() { return(speed_); }
		double& targetSpeed() { return(targetSpeed_); }
		int& targetSpeedChangeCount() { return(targetSpeedChangeCount_); }
		void updateSpeed(long&);
		void changeLane(double, double, double, double);
		// NodeList methods
		Node*& next() { return(next_); }
		Node*& prev() { return(prev_); }

	protected:
		int id_;
		double length_;
		Lane *lane_;
		double x_;
		double y_;
		double safeTime_;
		// IDM attributes
		double speed_;		// actual speed
		double targetSpeed_;	// desired speed
		int targetSpeedChangeCount_;
		int remainingBrakingSteps;
		int alreadyTraversed_;
		double dv;
		// NodeList attributes
		Node *next_;
		Node *prev_;
};

////////////////////////
// FIFO list of nodes //
////////////////////////
//
//  head         next         next         NULL
//  --->  +---+  --->  +---+  --->  +---+  --->
//        | A |        | B |        | C |
//  <---  +---+  <---  +---+  --->  +---+  <---
//  NULL         prev         prev         tail
//
//  with { x_A > x_B > x_B }
//
class NodeList {
	public:
		NodeList() { head_ = tail_ = NULL; }

		void  pushNode(Node *);             // push node into tail
		void  addNode(Node *);              // add the passed node to the list
		Node* popNode();                    // pop node from head
		void  delNode(Node *);              // remove the passed node fomr the list
		Node* headNode() { return(head_); } // access head node
		Node* tailNode() { return(tail_); } // access tail node

		// IDM methods
		Node* aheadNode(Node *); // access node ahead of given one
		Node* backNode(Node *);  // access node back of given one

	protected:
		Node *head_;
		Node *tail_;
};

///////////////
// Road lane //
////////////////
class Lane { 
	public:
		Lane(int id,int arrivalDistribution,
				double minSpeed, double maxSpeed,
				double acceleration, double deceleration,
				double jamDistance, double safeTime,
				int targetSpeedChangeMean,
				//double normalWeight,
				double mu , double sigma,
				//double g, double m,
				char *laneArrivalFile) : nodes_() {

			arrivalDistribution_ = arrivalDistribution;
			nextArrival_ = 0;
			nextSpeed_ = 0;
			nextIat_ = 0;
			pendingArrival_ = 0;
			
			totST_=0;
			sumST_=0.0;
			stmu_=stmu_lane[id];
			stsigma_=stsigma_lane[id];
			fprintf(stderr,"Lane %d with %.3f %.3f\n",id,stmu_,stsigma_);
			minSpeed_ = minSpeed;
			maxSpeed_ = maxSpeed;
			acceleration_ = acceleration;
			deceleration_ = deceleration;
			jamDistance_ = jamDistance;
			safeTime_ = safeTime;

			targetSpeedChangeMean_ = targetSpeedChangeMean;
			mu_=mu;
			sigma_=sigma;

			laneArrivalFileName_=laneArrivalFile;
			laneArrivalFile_ = fopen(laneArrivalFileName_, "r");
			printf("%s\n",laneArrivalFileName_);
			if (laneArrivalFile_ == NULL) {
				fprintf(stderr, "I couldn't open %s for reading.\n", laneArrivalFileName_);
				exit(0);
			}
		}

		int& arrivalDistribution() { return(arrivalDistribution_); }
		double& nextArrival() { return(nextArrival_); }
		double& nextSpeed() { return(nextSpeed_); }
		int& pendingArrival() { return(pendingArrival_); }

		double& minSpeed() { return(minSpeed_); }
		double& maxSpeed() { return(maxSpeed_); }
		double& acceleration() { return(acceleration_); }
		double& deceleration() { return(deceleration_); }
		double& jamDistance() { return(jamDistance_); }
		double& safeTime() { return(safeTime_); }

		int& targetSpeedChangeMean() { return(targetSpeedChangeMean_); }

		double& mu() { return(mu_);}
		double& sigma() { return(sigma_);}
		
		double& sumST() {return(sumST_);}
		int& totST() {return(totST_);}
		char*& laneArrivalFileName() {return(laneArrivalFileName_);}

		void readNextArrival(double currTime){
			double iat;
			double speedKmh;
			double time;
			int vid;
			if (fscanf(laneArrivalFile_, "%lf %lf %lf %d\n", &iat, &speedKmh, &time,&vid) == 4){
				nextArrival_=currTime + iat;
				nextSpeed_= speedKmh/3.6;
				nextIat_=iat;
			}else{

				nextArrival_= 1000000.0; //no more vehicles to enter this road
				iat=1000000.0;
				nextSpeed_= 100.0;
				vid=0;
				fprintf(stderr, "Error reading lane arrivals or EOF found\n");
			}
#if DEBUG
			fprintf(stderr,"lane %s iat: %lf\n",laneArrivalFileName_, iat);
#endif
			return;
		}

		double getTargetSpeed(double currSpeed){
			if (currSpeed > ((mu_+(2*sigma_))/3.6))
				return currSpeed;
			std::normal_distribution<double> distribution(mu_,sigma_);
			double candSpeed = 0.0;
			while (!(candSpeed>=currSpeed))
			{
				candSpeed = distribution(generator)/3.6;
			}

			return candSpeed;
		}
		double getSafeTime(){
			std::normal_distribution<double> distribution(stmu_,stsigma_);
			double candSt=100000.0;
			if (nextIat_ <= 0.2)
			{
				candSt=0.5;
			}
			else 
			{
				if (nextIat_ <= (stmu_-(2*stsigma_)))
				{
					candSt = nextIat_;
				}
				else
				{
					if (nextIat_ > 0)
					{
						while((candSt>nextIat_) || (candSt<0.3))
						{
							candSt=distribution(generator);
						}
					}
					else
					{
						candSt=0.65;
					}
				}
			}
			totST_++;
			sumST_+=candSt;
			return candSt;
		}
		NodeList& nodes() { return(nodes_); }



		// IDM methods
		double aheadSpeed(Node *);
		double aheadDistance(Node *);

	protected:
		int arrivalDistribution_;
		double nextArrival_;
		double nextSpeed_;
		double nextIat_;
		int pendingArrival_;

		// IDM attributes
		double minSpeed_;
		double maxSpeed_;
		double acceleration_;
		double deceleration_;
		double jamDistance_;
		double safeTime_;

		int targetSpeedChangeMean_;  


		double mu_;
		double sigma_;
		
		double stmu_;
		double stsigma_;
		
		int totST_;
		double sumST_;

		std::normal_distribution<double> distribution_;

		char *laneArrivalFileName_;
		FILE *laneArrivalFile_;

		NodeList nodes_;
};

//////////////////////
// Global functions //
//////////////////////
void computeNextArrival(int, long&);


//////////////////////
// Global variables //
//////////////////////
Lane **lanes;
Node **nodes;
NodeList inactiveNodes;
int numLanes;
double currTime;
double nextLogTime;
double timeStep;
FILE *log_fp;

///////////////////
// Main function //
///////////////////
int main(int argc, char **argv) {
	// mobility model-related variables
	double startX, startY, roadLength, laneWidth;
	double startTime, stopTime;
	int numNodes;
	double nodeLength;
	// internal variables
	Node *currNode;
	long seedArrival = 1;
	long seedTargetSpeed = 2;

	////////////////////
	// initialization //
	////////////////////
	lanes = NULL;
	nodes = NULL;
	if(argc < (1 + 11) ||
			argc != (1 + 11 + atoi(argv[1]) * 11)) {

		fprintf(stderr, "Usage: %s\n"
				"       <num_lanes> <startX> <startY>\n"
				"       <roadLength> <laneWidth>\n"
				"       <startTime> <stopTime> <numNodes>\n"
				"       <nodeLength> <timeStep> <outputFile>\n"
				"       { <arrivalDistribution> <arrivalFileName>\n"
				"       <minSpeed> <maxSpeed> <acceleration>\n"
				"       <deceleration> <jamDistance>\n"
				"       <safeTime> <speedChangeMean>\n"
				"       <mu> <sigma>}\n",
				argv[0]);
		exit(1);
	}

	numLanes = atoi(argv[1]);
	startX = atof(argv[2]);
	startY = atof(argv[3]);
	roadLength = atof(argv[4]);
	laneWidth = atof(argv[5]);
	startTime = atof(argv[6]);
	stopTime = atof(argv[7]);
	numNodes = atoi(argv[8]);
	nodeLength = atof(argv[9]);
	timeStep = atof(argv[10]);


	char *log_filename = new char[FILENAME];
	strcpy(log_filename, argv[11]);
	printf("%s\n",log_filename);
	if((log_fp = fopen(log_filename, "w")) == NULL) {
		fprintf(stderr, "Cannot open file %s\n", log_filename);
		exit(1);
	}
	setbuf(log_fp, 0);

	lanes = new Lane*[numLanes];

	for(int n=0; n<numLanes; n++)  {
		lanes[n] = new Lane(n,atoi(argv[12+n*11]),
				atof(argv[14+n*11]),
				atof(argv[15+n*11]),
				atof(argv[16+n*11]),
				atof(argv[17+n*11]),
				atof(argv[18+n*11]),
				atof(argv[19+n*11]),
				atof(argv[20+n*11]),
				atof(argv[21+n*11]),
				atof(argv[22+n*11]),
				argv[13+n*11]  );
	}

	nodes = new Node*[numNodes];
	int nodeID = 0;

	// initialize off-road nodes
	for(int i=0; i<numNodes; i++) {
		currNode = new Node(nodeID++, nodeLength, NULL, OUT_X, OUT_Y);

		// store node in inactive list
		inactiveNodes.pushNode(currNode);

		// store node in node archive
		nodes[nodeID-1] = currNode;
	}

	// init time
	currTime = startTime;
	nextLogTime=0.0;

	// compute initial next arrival times for each lane
	for(int n=0; n<numLanes; n++)
		computeNextArrival(n, seedArrival);

	///////////////
	// main loop //
	///////////////
	while(currTime < stopTime) {

		unsigned int currStepNo = round(currTime / timeStep);
		if (currStepNo % 100 == 0)
			printf("%.3f\r",currTime);
		
		// handle nodes which arrived to the end of lane
		for(int n=0; n<numLanes; n++) {
			currNode = lanes[n]->nodes().headNode();

			while(currNode != NULL && currNode->x() == startX + roadLength) {

				// store the next node in the lane
				Node *nextNode = lanes[n]->nodes().backNode(currNode);

				// reset current node internal state
				currNode = lanes[n]->nodes().popNode();
#if DEBUG>1
				fprintf(stderr, "finished node: %d\n", currNode->id());
#endif
				currNode->lane() = NULL;
				currNode->x() = OUT_X;
				currNode->y() = OUT_Y;
				inactiveNodes.pushNode(currNode);

				// update next node to be processed
				currNode = nextNode;
			}
		}

		// process active nodes: lane changing
		for(int id=0; id<numNodes; id++) {
			currNode = nodes[id];

			if(currNode->y() != OUT_Y)
				currNode->changeLane(startX, startY, roadLength, laneWidth);
		}

		// process active nodes: acceleration
		for(int n=0; n<numLanes; n++) {
			currNode = lanes[n]->nodes().headNode();

			while(currNode != NULL){
				// IDM
				double prevSpeed = currNode->speed();
				currNode->updateSpeed(seedTargetSpeed);
				double movement = 0.5*(prevSpeed+currNode->speed())*timeStep;
				double remaining = startX + roadLength - currNode->x();

				// if lane end will be reached within this
				// time step, set speed and handle node arrival
				if(movement >= remaining) {
					currNode->x() = startX + roadLength;
					double arrivalTime = currTime + remaining / currNode->speed();

				} else {         // otherwise, proceed at computed speed
					currNode->x() += movement;
					
					// log mobility output          
					if (currStepNo % 5 == 0)
						fprintf(log_fp, "%.1f %d %.2f %.2f %.2f\n",
								currTime,currNode->id(),
								currNode->x(), currNode->y(),
								currNode->speed()*3.6);

				}

				currNode = lanes[n]->nodes().backNode(currNode);
			}
		}
		for(int n=0; n<numLanes; n++) { // handle new nodes starting on each lane
			// if an arrival must be processed

			unsigned int currStepNo = round(currTime / timeStep);
			unsigned int naStepNo = round(lanes[n]->nextArrival() / timeStep); 
			if(naStepNo <= currStepNo || lanes[n]->pendingArrival() > 0) {

				if(lanes[n]->nodes().tailNode() == NULL ||
						lanes[n]->nodes().tailNode()->x() - startX > lanes[n]->jamDistance()) {
					currNode = inactiveNodes.popNode();
					if(currNode == NULL) {
						fprintf(stderr, "Not enough nodes!\n");
						exit(0);
					}
					currNode->lane() = lanes[n];
					currNode->x() = startX;
					currNode->y() = startY + laneWidth * n;
					currNode->speed() = lanes[n]->nextSpeed();
					currNode->targetSpeed() =lanes[n]->getTargetSpeed(currNode->speed());
					currNode->targetSpeedChangeCount() = lanes[n]->targetSpeedChangeMean();
					currNode->safeTime() = lanes[n]->getSafeTime();
#if DEBUG
					if(lanes[n]->nodes().tailNode() != NULL)
						fprintf(stderr, "curr time %lf tail node %d at %lf %lf speed %lf target speed %lf\n",
								currTime,
								lanes[n]->nodes().tailNode()->id(),
								lanes[n]->nodes().tailNode()->x(),
								lanes[n]->nodes().tailNode()->y(),
								lanes[n]->nodes().tailNode()->speed(),
								lanes[n]->nodes().tailNode()->targetSpeed());
					fprintf(stderr, "curr time %lf new node %d at %lf %lf speed %lf target speed %lf\n",
							currTime,
							currNode->id(),
							currNode->x(),
							currNode->y(),
							currNode->speed(),
							currNode->targetSpeed());

#endif
					lanes[n]->nodes().pushNode(currNode);

					if(naStepNo <= currStepNo)
						computeNextArrival(n, seedArrival);
#if DEBUG
					fprintf(stderr, "Lane %d nextArrival: %lf\n",n, lanes[n]->nextArrival());
#endif

					// update the number of nodes waiting to start
					if(lanes[n]->pendingArrival() > 0)
						lanes[n]->pendingArrival()--;

			}else if(naStepNo <= currStepNo){
				fprintf(stderr, "waiting node, we shouldn't have of those!!\n");
				fprintf(stderr, "time %lf crashing with node %d at %lf %lf speed %lf\n",
						currTime,
						lanes[n]->nodes().tailNode()->id(),
						lanes[n]->nodes().tailNode()->x(),
						lanes[n]->nodes().tailNode()->y(),
						lanes[n]->nodes().tailNode()->speed());

				lanes[n]->pendingArrival()++;
				exit(1);
			}
			}
		} // end of new nodes handling

		// update time
		currTime += timeStep;

		} // end of main loop
		fprintf(stderr,"Total number of vehicles with deceleration < -max deceleration : %ld\n",maxDecelVSet.size());
		
		for(int n=0;n<numLanes;n++)
			printf("Avg st %.3f\n",lanes[n]->sumST()/lanes[n]->totST());
		fclose(log_fp);

		return(0);  
	}

	//////////////////////////////////////////////
	// Compute next arrival times for each lane //
	//////////////////////////////////////////////
	void computeNextArrival(int n, long& seedArrival) {
		switch(lanes[n]->arrivalDistribution()) {
			//case 0:  // deterministic arrivals
			//         lanes[n]->nextArrival() = currTime +
			//                                   lanes[n]->arrivalMean();
			//         break;
			//case 1:  // uniform arrivals
			//        lanes[n]->nextArrival() = currTime +
			//                                   uniform(0, (lanes[n]->arrivalMean())*2, &seedArrival);
			//         break;
			//case 2:  // exponential arrivals
			//         lanes[n]->nextArrival() = currTime +
			//                                   negexp(lanes[n]->arrivalMean(), &seedArrival);
			//         break;
			//case 3:
			//		 lanes[n]->nextArrival() = currTime +
			//                                   weightedNormalExponential(lanes[n]->normalWeight(),
			//                                   lanes[n]->mu(), lanes[n]->sigma(),
			//                                   lanes[n]->g(), lanes[n]->m());
			//		 break;
			case 4:
				lanes[n]->readNextArrival(currTime);
				//fprintf(log_fp, "lane %d, currTime: %lf next arrival: %lf (%lf) speed: %lf %lf\n",
				//					n,
				//					currTime,
				//					lanes[n]->nextArrival(),
				//					lanes[n]->nextArrival()-currTime,
				//					lanes[n]->nextSpeed(),
				//					lanes[n]->nextSpeed()*3.6);
				break;
			default:fprintf(stderr, "Unknown arrival distribution for lane %d: %d\n", n,lanes[n]->arrivalDistribution());
				exit(1);
		}

		return;
	}


	float weightedNormalExponential(float normalWeight, float mu, float sigma, float g, float m){
		std::normal_distribution<> norm_dist(mu,sigma);
		std::exponential_distribution<> exp_dist(g);
		return (normalWeight*norm_dist(gen)+ (1-normalWeight)*(exp_dist(gen)+m));
	}



	////////////////////////////////
	// Get speed of ahead vehicle //
	////////////////////////////////
	double Lane::aheadSpeed(Node* node) {
		Node* currNode = NULL;

		if(node == nodes_.headNode())
			return(0.0);

		currNode = node->prev();
		if(currNode == NULL) {
			fprintf(stderr, "Cannot find ahead node speed\n");
			exit(1);
		}

		return(currNode->speed());
	}

	/////////////////////////////////////////
	// Compute distance from ahead vehicle //
	/////////////////////////////////////////
	double Lane::aheadDistance(Node* node) {
		Node* currNode = NULL;

		if(nodes_.aheadNode(node) == NULL)
			return(MAX_DOUBLE);

		currNode = node->prev();
		if(currNode == NULL) {
			fprintf(stderr, "Cannot find ahead node distance\n");
			exit(1);
		}

		// x identifies the end of vehicle, must account for length
		double distance = currNode->x() - node->x() - node->length();

		// check if the distance is lower than zero
		if(distance <= 0) {
		
			fprintf(stderr, "Negative distance %f between nodes %d (x=%f) and %d (x=%f)\n"
					"with pending arrival %d and tail node %d at lane %s\n",
					distance, node->id(), node->x(), currNode->id(), currNode->x(),
					node->lane()->pendingArrival(), node->lane()->nodes().tailNode()->id(),
					node->lane()->laneArrivalFileName());
			fprintf(stderr,"Node %d speed %.3f\n",currNode->id(),currNode->speed());
			exit(1);
	
		}

		return(distance);
	}

	


	/////////////////////////////////
	// Update node's current speed //
	/////////////////////////////////
	void Node::updateSpeed(long& seedTargetSpeed) {

		// update targetSpeed every [targetSpeedChangeCount_*timeStep]:
		//  - new target speed value is smoothed over current one
		//  - the time interval between speed updates is deterministic,
		//    however a random distribution could be implemented

		
		std::uniform_real_distribution<> dis(0,1);
		double speedDifference = speed_ - lane_->aheadSpeed(this);
		double distance = lane_->aheadDistance(this);
		// safety distance check
		if (distance <= 0) {
			speed_ = 0.0;
			return;
		}

		
		double add = speed_*safeTime_ +
				speed_*speedDifference/(2.0*sqrt(lane_->acceleration()*lane_->deceleration()));
		if(add < 0)
			add = 0;
		double targetDistance = lane_->jamDistance() + add;

		dv = lane_->acceleration() *
			(1 - pow(speed_/targetSpeed_, 4) - pow(targetDistance/distance, 2));

		if (dv < -MAXDECEL)
		{
			if (maxDecelVSet.count(this->id())==0)
				maxDecelVSet.insert(this->id());
			dv = -MAXDECEL;
		}

		speed_ += dv * timeStep;
		// reach desired speed
		if(speed_ > targetSpeed_)
			speed_ = targetSpeed_;


		// do a traffic jam
		if(isnan(speed_) || speed_ < 0.0)
			speed_ = 0.0;
	}

	///////////////////////////////////
	// Check if node can change lane //
	///////////////////////////////////
	void Node::changeLane(double startX, double startY, double roadLength, double laneWidth) {
		int curLane, newLane, move;
		double rightAdv, leftAdv;
		double bSave, gapMin, rightLaneBias, politeness, thresholdAdv;
		double newBackNewSpeedDiff, newBackNewDist, newBackNewAcc;
		double newBackCurSpeedDiff, newBackCurDist, newBackCurAcc;
		double curBackNewSpeedDiff, curBackNewDist, curBackNewAcc;
		double curBackCurSpeedDiff, curBackCurDist, curBackCurAcc;
		double thisNewSpeedDiff, thisNewDist, thisNewAcc;
		double thisCurSpeedDiff, thisCurDist, thisCurAcc;
		double add, targetDistance, disadvantage;
		Node *newBackNode, *curBackNode, *newFrontNode;

		newBackNode = curBackNode = newFrontNode = NULL;

		// define paramaters
		bSave = lane_->deceleration() + 1.0;            // safe breaking deceleration (m/s^2)
		//gapMin = length_ + lane_->safeTime() * speed_;  // minimum safe distance (m)
		gapMin = length_ + safeTime_ * speed_;
		//rightLaneBias = 0.2;                            // favour right lane (m/s^2)(default 0.2)
		rightLaneBias = RIGHTLANEBIAS;
		//END PATCH
		politeness = 0.5;                               // politeness factor (1 = very polite, 0 = very selfish) (default 0.5)
		//thresholdAdv = 0.2;                             // minimum advantage needed to change lane(default 0.4)
		thresholdAdv = THRESHOLDADV;
		// find which lane the node is travelling on
		
		
		
		for(curLane=0; curLane<numLanes; curLane++) {
			if(lanes[curLane] == lane_)
				break;
		}
		if(curLane >= numLanes) {
			fprintf(stderr, "   could not find node's lane\n");
			exit(1);
		}

		// evaluate a possible move to right (-1) or left (+1) lane
		leftAdv = rightAdv = 0;
		for(move=-1; move<=1; move+=2) {
			newLane = curLane+move;

			// check if new lane exists
			if((move == 1  && newLane <  numLanes) ||
					(move == -1 && newLane >= 0)) {
				// find back node in other lane
				newBackNode = lanes[newLane]->nodes().headNode();
				while(newBackNode != NULL) {
					if(newBackNode->x() < x_)
						break;
					newBackNode = lanes[newLane]->nodes().backNode(newBackNode);
				}

				// if back node exists, find front node
				// and calculate its new and current accelerations
				if(newBackNode != NULL) {

					newFrontNode = lanes[newLane]->nodes().aheadNode(newBackNode);

					// compute back node new acceleration using IDM formulae
					newBackNewSpeedDiff = newBackNode->speed() - speed_;
					newBackNewDist = x_ - newBackNode->x() - newBackNode->length();
					// do not overpass if safety conditions are not verified
					if(newBackNewDist < gapMin) {
						continue; 
					}
					
					add = speed_*safeTime_ + newBackNode->speed()*newBackNewSpeedDiff/
						(2.0*sqrt(lanes[newLane]->acceleration()*lanes[newLane]->deceleration()));
					if(add < 0)
						add = 0;
					targetDistance = lanes[newLane]->jamDistance() + add;
					
					newBackNewAcc = lanes[newLane]->acceleration() *
                                              (1 - pow(newBackNode->speed()/newBackNode->targetSpeed(), 4) - pow(targetDistance/newBackNewDist, 2));
					// compute back node current acceleration using IDM formulae
					newBackCurSpeedDiff = newBackNode->speed() - lanes[newLane]->aheadSpeed(newBackNode);
					newBackCurDist = lanes[newLane]->aheadDistance(newBackNode);
					
					add = newBackNode->speed()*newBackNode->safeTime() + newBackNode->speed()*newBackCurSpeedDiff/
						(2.0*sqrt(lanes[newLane]->acceleration()*lanes[newLane]->deceleration()));
					if(add < 0)
						add = 0;
					targetDistance = lanes[newLane]->jamDistance() + add;
					
					
					newBackCurAcc = lanes[newLane]->acceleration() *
						(1 - pow(newBackNode->speed()/newBackNode->targetSpeed(), 4) - pow(targetDistance/newBackCurDist, 2));
				
				}

				// if back node does not exist, compute distance from road start,
				// find first node and set newBackNode new and current acceleration to zero
				else {
					newBackNewDist = x_ - startX;
					// do not overpass if safety conditions are not verified
					if(newBackNewDist < gapMin) {
						continue;
					}
					newFrontNode = lanes[newLane]->nodes().tailNode();
					newBackNewAcc = 0.0;
					newBackCurAcc = 0.0;
				}

				// do not overpass if safety conditions are not verified
				if(newBackNewAcc < -bSave) {
					continue;
				}

				// set up speed difference and distance with respect
				// to front node when front node exists 
				if(newFrontNode != NULL) {

					thisNewSpeedDiff = speed_ - newFrontNode->speed();
					thisNewDist = newFrontNode->x() - x_ - length_;
				}
				// set up speed difference and distance with respect
				// to front node when front node does not exist 
				else {
					thisNewSpeedDiff = speed_ - 0.0;
					thisNewDist = MAX_DOUBLE;
				}

				// do not overpass if safety conditions are not verified
				if(thisNewDist < gapMin) {
					continue;
				}

				// compute new acceleration using IDM formulae

				add = speed_*safeTime_ + speed_*thisNewSpeedDiff/
					(2.0*sqrt(lanes[newLane]->acceleration()*lanes[newLane]->deceleration()));
				if(add < 0)
					add = 0;
				targetDistance = lanes[newLane]->jamDistance() + add;
				thisNewAcc = lanes[newLane]->acceleration() *
					(1 - pow(speed_/targetSpeed_, 4) - pow(targetDistance/thisNewDist, 2));

				// compute current acceleration using IDM formulae
				thisCurSpeedDiff = speed_ - lane_->aheadSpeed(this);
				thisCurDist = lane_->aheadDistance(this);

				add = speed_*safeTime_ + speed_*thisCurSpeedDiff/
					(2.0*sqrt(lane_->acceleration()*lane_->deceleration()));
				if(add < 0)
					add = 0;
				targetDistance = lane_->jamDistance() + add;
				thisCurAcc = lane_->acceleration() *
					(1 - pow(speed_/targetSpeed_, 4) - pow(targetDistance/thisCurDist, 2));

				// compute own advantage in changing lane
				double advantage= thisNewAcc - thisCurAcc;
				if(move == -1)
					advantage += thisNewAcc - thisCurAcc + rightLaneBias;

				// compute others disadvantage in changing lane
				disadvantage = newBackCurAcc - newBackNewAcc;

				// if moving to the right, we must also account for 'pushy'
				// vehicles moving at higher speed back of the current vehicle
				if(move == -1) {
					curBackNode = lane_->nodes().backNode(this);

					// if back node exists, find front node
					// and calculate its new and current accelerations
					if(curBackNode != NULL) {
						// compute back node new acceleration using IDM formulae
						curBackNewSpeedDiff = curBackNode->speed() - lane_->aheadSpeed(this);
						curBackNewDist = (lane_->aheadDistance(this) - length_) + (x_ - curBackNode->x() - curBackNode->length());

						add = curBackNode->speed()*curBackNode->safeTime() + curBackNode->speed()*curBackNewSpeedDiff/
							(2.0*sqrt(lane_->acceleration()*lane_->deceleration()));
						if(add < 0)
							add = 0;
						targetDistance = lane_->jamDistance() + add;

					        newBackNewAcc = lane_->acceleration() *
                                                        (1 - pow(curBackNode->speed()/curBackNode->targetSpeed(), 4) - pow(targetDistance/curBackNewDist, 2));				
						// compute back node current acceleration using IDM formulae
						curBackCurSpeedDiff = curBackNode->speed() - speed_;
						curBackCurDist = x_ - curBackNode->x() - curBackNode->length();

						add = curBackNode->speed()*curBackNode->safeTime() + curBackNode->speed()*curBackCurSpeedDiff/
							(2.0*sqrt(lanes[newLane]->acceleration()*lanes[newLane]->deceleration()));
						if(add < 0)
							add = 0;
						targetDistance = lanes[newLane]->jamDistance() + add;

						curBackCurAcc = lanes[newLane]->acceleration() *
                                                        (1 - pow(curBackNode->speed()/curBackNode->targetSpeed(), 4) - pow(targetDistance/curBackCurDist, 2));
					}
					// if back node does not exist, compute distance from road start,
					// find first node and set newBackNode new and current acceleration to zero
					else {
						curBackNewAcc = 0.0;
						curBackCurAcc = 0.0;
					}

					disadvantage += curBackCurAcc - curBackNewAcc;
				}

				// disadvantage cannot be lower than zero
				if(disadvantage < 0)
					disadvantage = 0;

				// compute 'net' advantage
				advantage -= politeness*disadvantage;

				// store computed advantage according to the movement
				if(move == -1)
					rightAdv = advantage;
				if(move == 1)
					leftAdv = advantage;
			}
		}

		// move to right lane
		if(rightAdv >= leftAdv &&
				rightAdv > thresholdAdv) {
			lane_ = lanes[curLane-1];
			y_ = startY + laneWidth * (curLane-1);
			lanes[curLane]->nodes().delNode(this);
			lanes[curLane-1]->nodes().addNode(this);

			return;
		}

		// move to left lane
		if(leftAdv > rightAdv &&
				leftAdv > thresholdAdv) {
			lane_ = lanes[curLane+1];
			y_ = startY + laneWidth * (curLane+1);
			lanes[curLane]->nodes().delNode(this);
			lanes[curLane+1]->nodes().addNode(this);

			return;
		}

		// do not change lane
		return;
	}

	void NodeList::pushNode(Node *node) {
		node->prev() = NULL;
		node->next() = NULL;

		if(head_ == NULL) {
			head_ = tail_ = node;
		}
		else {
			tail_->next() = node;
			node->prev() = tail_;
			tail_ = node;
		}

		return;
	}

	void NodeList::addNode(Node *newNode) {
		Node *currNode;

		// setup new node
		newNode->prev() = NULL;
		newNode->next() = NULL;

		// empty list
		if(head_ == NULL) {
			head_ = tail_ = newNode;
		}
		// non-empty list
		else {
			currNode = head_;
			while(currNode != NULL) {
				if(currNode->x() < newNode->x())
					break;
				currNode = backNode(currNode);
			}
			// head insertion
			if(currNode == head_) {
				currNode->prev() = newNode;
				newNode->next() = currNode;
				head_ = newNode;
			}
			// tail insertion
			else if(currNode == NULL) {
				tail_->next() = newNode;
				newNode->prev() = tail_;
				tail_ = newNode;
			}    
			// normal insertion
			else {
				currNode->prev()->next() = newNode;
				newNode->prev() = currNode->prev();
				currNode->prev() = newNode;
				newNode->next() = currNode;
			}
		}

		return;
	}

	Node* NodeList::popNode() {
		Node *node = NULL;

		if(head_ != NULL) {
			node = head_;
			head_ = node->next();
			node->next() = NULL;
			if(head_ != NULL)
				head_->prev() = NULL;
		}

		return(node);
	}

	void NodeList::delNode(Node *oldNode) {
		Node *currNode;

		// empty list
		if(head_ == NULL) {
			fprintf(stderr, "Deleting a node from an empty list\n");
			exit(1);
		}
		// non-empty list
		else {
			currNode = head_;
			while(currNode != NULL) {
				if(currNode == oldNode)
					break;
				currNode = backNode(currNode);
			}
			// could not find node
			if(currNode == NULL) {
				fprintf(stderr, "Deleting non-existent node from list\n");
				exit(1);      
			}    
			// head deletion
			if(currNode == head_) {
				head_ = currNode->next();
				if(head_ != NULL)
					head_->prev() = NULL;
				currNode->prev() = NULL;
				currNode->next() = NULL;
			}
			// tail deletion
			else if(currNode == tail_) {
				tail_ = currNode->prev();
				if(tail_ != NULL)
					tail_->next() = NULL;
				currNode->prev() = NULL;
				currNode->next() + NULL;
			}
			// normal deletion
			else {
				currNode->next()->prev() = currNode->prev();
				currNode->prev()->next() = currNode->next();
				currNode->prev() = NULL;
				currNode->next() + NULL;      
			}
		}

		return;
	}

	Node* NodeList::aheadNode(Node *node) {
		return(node->prev());
	}

	Node* NodeList::backNode(Node *node) {
		return(node->next());
	}
